@@ -302,7 +302,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		sdl_type = 0;
 		/* Don't allow loopback unless explicit */
 		if (ifp->flags & IFF_LOOPBACK) {
-			if (argc == 0 && ctx->ifac == 0) {
+			if ((argc == 0 || argc == -1) && ctx->ifac == 0) {
 				if_free(ifp);
 				continue;
 			}
@@ -326,11 +326,22 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 			}
 #endif
 
+#ifdef __FreeBSD__
+			memcpy(&ifp->linkaddr, sdl, sdl->sdl_len);
+#endif
 			ifp->index = sdl->sdl_index;
 			sdl_type = sdl->sdl_type;
 			switch(sdl->sdl_type) {
 #ifdef IFT_BRIDGE
-			case IFT_BRIDGE: /* FALLTHROUGH */
+			case IFT_BRIDGE:
+				/* Don't allow bridge unless explicit */
+				if ((argc == 0 || argc == -1)
+				    && ctx->ifac == 0)
+				{
+					if_free(ifp);
+					continue;
+				}
+				/* FALLTHOUGH */
 #endif
 #ifdef IFT_L2VLAN
 			case IFT_L2VLAN: /* FALLTHOUGH */
@@ -376,7 +387,7 @@ if_discover(struct dhcpcd_ctx *ctx, int argc, char * const *argv)
 		if (!(ifp->flags & IFF_POINTOPOINT) &&
 		    ifp->family != ARPHRD_ETHER)
 		{
-			if (argc == 0 && ctx->ifac == 0) {
+			if ((argc == 0 || argc == -1) && ctx->ifac == 0) {
 				if_free(ifp);
 				continue;
 			}
